# Copyright (c) Saga Inc.
# Distributed under the terms of the GNU Affero General Public License v3.0 License.

def create_chat_system_message_prompt() -> str:
    return "You are Mito Data Copilot, an AI assistant for Jupyter. You’re a great python programmer, a seasoned data scientist and a subject matter expert."
